# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2019. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************

# Sample program that uses cuml with dask to find nearest neighbors

import os
import numpy as np
import pandas as pd
import argparse
import logging
CLI=argparse.ArgumentParser()
CLI.add_argument(
    "--nrows",
    type=int,
    default=50000
)
CLI.add_argument(
    "--ncols",
    type=int,
    default=100
)
CLI.add_argument(
    "--n_neighbors",
    type=int,
    default=5
)
CLI.add_argument(
    "--n_workers",
    type=int
)
CLI.add_argument(
    "--dask_scheduler",
    type=str
)
CLI.add_argument(
    "--scheduler_port",
    type=int,
    default=0
)

args = CLI.parse_args()

# Create pandas dataframe with random data
def load_data(nrows, ncols):
    print('Generating random data')
    X = np.random.random((nrows,ncols)).astype('float32')
    df = pd.DataFrame({'fea%d'%i:X[:,i] for i in range(X.shape[1])}).fillna(0.0)
    return df

if __name__ == '__main__':

    if (args.dask_scheduler is None):
        # If user did not pass dask-scheduler URL through --dask_scheduler option,
        # create a local Dask CUDA cluster with one worker per device.
        if (args.scheduler_port == 0):
            print("--scheduler_port not passed. Scheduler will use random port")

        from dask_cuda import LocalCUDACluster
        if (args.n_workers is None):
            print("--n_workers option not passed. Using all GPUs...")
            cluster = LocalCUDACluster(scheduler_port=args.scheduler_port)
        else:
            cluster = LocalCUDACluster(n_workers=args.n_workers,scheduler_port=args.scheduler_port)
        print("Cluster status: ", cluster.status)
    else:
        cluster = args.dask_scheduler

    from distributed import Client
    client = Client(cluster)

    print("Cluster information: ", cluster)
    print("Client information: ", client)
    n_workers = len(client.has_what())
    print('num_workers:', n_workers)

    X = load_data(args.nrows, args.ncols)
    print('data', X.shape)
    import cudf
    cdf = cudf.DataFrame.from_pandas(X)

    pdf = cdf.to_pandas()

    import dask.dataframe as dd
    df = dd.from_pandas(pdf, npartitions=n_workers)

    import dask_cudf
    ddf = dask_cudf.from_dask_dataframe(df)
    #print(ddf.shape)

    # Fit a NearestNeighbors model and query it
    from cuml.dask.neighbors import NearestNeighbors
    nn = NearestNeighbors(n_neighbors = args.n_neighbors)
    nn.fit(ddf)

    # Query distances (D) to the neighbors of each point and indices (I) of
    # the neighbors of each point in ddf.
    D, I = nn.kneighbors(ddf)

    I_nd = np.array(I.compute().as_gpu_matrix(), dtype='int64')
    D_nd = np.array(D.compute().as_gpu_matrix(), dtype='float32')

    I_nd = np.sort(I_nd, axis=0)
    D_nd = np.sort(D_nd, axis=0)

    print("[dask-cuml]: Distances to the neighbors of each point:", D_nd)
    print("[dask_cuml]: Indices of the neighbors of each point:\n", I_nd)
    if (args.dask_scheduler is None):
        cluster.close()

